#include "ASXTransform.h"

CASXTransform::CASXTransform(void)
{
	nTransformLength = 1024;
	NumBits = NumberOfBitsNeeded();
}

unsigned int CASXTransform::NumberOfBitsNeeded()
{
	int i;

	if( nTransformLength < 2 )
	{
		return 0;
	}

	for ( i=0; ; i++ )
	{
		if( nTransformLength & (1 << i) ) return i;
    }
}

unsigned int CASXTransform::ReverseBits(unsigned int p_nIndex, unsigned int p_nBits)
{
	unsigned int i, rev;

	for(i=rev=0; i < p_nBits; i++)
	{
		rev = (rev << 1) | (p_nIndex & 1);
		p_nIndex >>= 1;
	}

	return rev;
}

double CASXTransform::Index_to_frequency( unsigned NumSamples, unsigned Index )
{
    if ( Index >= NumSamples )
        return 0.0;
    else if ( Index <= NumSamples/2 )
        return (double)Index / (double)NumSamples;

    return -(double)(NumSamples-Index) / (double)NumSamples;
}

bool CASXTransform::IsPowerOfTwo( unsigned int p_nX )
{
	if( p_nX < 2 ) return false;
	if( p_nX & (p_nX-1) ) return false;
    return true;
}

CASXTransform::~CASXTransform(void)
{
}


CASXFFT::CASXFFT(void)
{
	nTransformLength = 1024;
	NumBits = NumberOfBitsNeeded();
}

CASXFFT::~CASXFFT(void)
{
}

// Fast Fourier Transform
void CASXFFT::fft(double *p_RealIn, double *p_RealOut, double *p_ImagOut)
{
	unsigned int i, j, k, n;
	unsigned int BlockSize, BlockEnd;

	double angle_numerator = 2.0 * PI;
	double tr, ti;

	if( !IsPowerOfTwo(nTransformLength) )
	{
		return;
	}

	for( i=0; i < nTransformLength; i++ )
	{
		j = ReverseBits ( i, NumBits );
		p_RealOut[j] = p_RealIn[i];
		p_ImagOut[j] = 0.0;
	}

	BlockEnd = 1;
	for( BlockSize = 2; BlockSize <= nTransformLength; BlockSize <<= 1 )
	{
		double delta_angle = angle_numerator / (double)BlockSize;
		double sm2 = sin ( -2 * delta_angle );
		double sm1 = sin ( -delta_angle );
		double cm2 = cos ( -2 * delta_angle );
		double cm1 = cos ( -delta_angle );
		double w = 2 * cm1;
		double ar[3], ai[3];

		for( i=0; i < nTransformLength; i += BlockSize )
		{

			ar[2] = cm2;
			ar[1] = cm1;

			ai[2] = sm2;
			ai[1] = sm1;

			for ( j=i, n=0; n < BlockEnd; j++, n++ )
			{

				ar[0] = w*ar[1] - ar[2];
				ar[2] = ar[1];
				ar[1] = ar[0];

				ai[0] = w*ai[1] - ai[2];
				ai[2] = ai[1];
				ai[1] = ai[0];

				k = j + BlockEnd;
				tr = ar[0]*p_RealOut[k] - ai[0]*p_ImagOut[k];
				ti = ar[0]*p_ImagOut[k] + ai[0]*p_RealOut[k];

				p_RealOut[k] = p_RealOut[j] - tr;
				p_ImagOut[k] = p_ImagOut[j] - ti;

				p_RealOut[j] += tr;
				p_ImagOut[j] += ti;
			}
		}

		BlockEnd = BlockSize;
	}
}

// Inverse Fast Fourier Transform
void CASXFFT::ifft(double *p_lpRealIn, double *p_lpImagIn, double *p_lpRealOut)
{
	unsigned int i, j, k, n;
	unsigned int BlockSize, BlockEnd;

	double angle_numerator = 2.0 * PI;
	double tr, ti;

	if( !IsPowerOfTwo(nTransformLength) )
	{
		return;
	}

	angle_numerator = -angle_numerator;

	for( i=0; i < nTransformLength; i++ )
	{
		j = ReverseBits ( i, NumBits );
		p_lpRealOut[j] = p_lpRealIn[i];
		//p_lpImagIn[j] = p_lpImagIn[i];
	}

	for(int i=0; i < nTransformLength/2; i++)
	{
		j = ReverseBits ( i, NumBits );
		double temp = p_lpImagIn[j];
		p_lpImagIn[j] = p_lpImagIn[i];
		p_lpImagIn[i] = temp;
	}


	BlockEnd = 1;
	for( BlockSize = 2; BlockSize <= nTransformLength; BlockSize <<= 1 )
	{
		double delta_angle = angle_numerator / (double)BlockSize;
		double sm2 = sin ( -2 * delta_angle );
		double sm1 = sin ( -delta_angle );
		double cm2 = cos ( -2 * delta_angle );
		double cm1 = cos ( -delta_angle );
		double w = 2 * cm1;
		double ar[3], ai[3];

		for( i=0; i < nTransformLength; i += BlockSize )
		{

			ar[2] = cm2;
			ar[1] = cm1;

			ai[2] = sm2;
			ai[1] = sm1;

			for ( j=i, n=0; n < BlockEnd; j++, n++ )
			{

				ar[0] = w*ar[1] - ar[2];
				ar[2] = ar[1];
				ar[1] = ar[0];

				ai[0] = w*ai[1] - ai[2];
				ai[2] = ai[1];
				ai[1] = ai[0];

				k = j + BlockEnd;
				tr = ar[0]*p_lpRealOut[k] + ai[0]*p_lpImagIn[k];
				ti = ar[0]*p_lpImagIn[k] + ai[0]*p_lpRealOut[k];

				p_lpRealOut[k] = p_lpRealOut[j] - tr;
				p_lpImagIn[k] = p_lpImagIn[j] - ti;

				p_lpRealOut[j] += tr;
				p_lpImagIn[j] += ti;
			}
		}
		BlockEnd = BlockSize;
	}

	double denom = (double)nTransformLength;

	for ( i=0; i < nTransformLength; i++ )
	{
		p_lpRealOut[i] /= denom;
	}
}